from views_pipeline_core.managers.model import ForecastingModelManager
from views_stepshifter.models.hurdle_model import HurdleModel
from views_stepshifter.models.validation import views_validate
from sklearn.utils.validation import check_is_fitted
import pandas as pd
from typing import List, Dict
import numpy as np
import logging
from tqdm import tqdm

logger = logging.getLogger(__name__)


class ShurfModel(HurdleModel):
    def __init__(self, config: Dict, partitioner_dict: Dict[str, List[int]]):
        super().__init__(config, partitioner_dict)
        self._clf_params = self._get_parameters(config)["clf"]
        self._reg_params = self._get_parameters(config)["reg"]

        self._submodel_list = []
        self._submodels_to_train = config["submodels_to_train"]
        self._log_target = config["log_target"]
        self._pred_samples = config["pred_samples"]
        self._draw_dist = config["draw_dist"]
        self._draw_sigma = config["draw_sigma"]

        # self._n_estimators = config['parameters']['n_estimators']
        # self._max_features = config['max_features']
        # self._max_depth = config['max_depth']
        # self._max_samples = config['max_samples']
        # self._geo_unit_samples = config['geo_unit_samples']
        # self._n_jobs = config['n_jobs']

    @views_validate
    def fit(self, df: pd.DataFrame):
        """
        Generate predictions using the trained submodels.
        This method performs the following steps:
        1. Prepares the data for classification and regression stages.
        2. Iterates over each submodel to generate predictions:
            - Predicts probabilities using the classification model.
            - Predicts target values using the regression model.
            - Handles infinite values in predictions.
        3. Draws samples from the distributions:
            - For each prediction sample, combines classification and regression predictions.
            - Applies binomial, Poisson, or lognormal distributions to generate final predictions.
        4. Aggregates the predictions from all submodels into a final DataFrame.
        Returns:
            pd.DataFrame: A DataFrame containing the final set of predictions with indices set to 'draw'.
        """
        df = self._process_data(df)
        self._prepare_time_series(df)
        self._clf = self._resolve_clf_model(self._config["model_clf"])
        self._reg = self._resolve_reg_model(self._config["model_reg"])

        target_binary = [
            s.map(lambda x: (x > 0).astype(float)) for s in self._target_train
        ]

        target_pos, past_cov_pos = zip(
            *[
                (t, p)
                for t, p in zip(self._target_train, self._past_cov)
                if (t.values() > 0).any()
            ]
        )

        for i in tqdm(range(self._submodels_to_train), desc="Training submodel"):

            for step in tqdm(self._steps, desc=f"Steps for submodel {i+1}"):
                # Fit binary-like stage using a regression model, but the target is binary (0 or 1)
                binary_model = self._clf(
                    lags_past_covariates=[-step], **self._clf_params
                )
                binary_model.fit(target_binary, past_covariates=self._past_cov)

                # Fit positive stage using the regression model
                positive_model = self._reg(
                    lags_past_covariates=[-step], **self._reg_params
                )
                positive_model.fit(target_pos, past_covariates=past_cov_pos)
                self._models[step] = (binary_model, positive_model)

            self._submodel_list.append(self._models)

        self.is_fitted_ = True

    def predict_sequence(self, sequence_number) -> pd.DataFrame:
        """
        Predicts n draws of outcomes based on the provided DataFrame .

        Parameters:
        -----------
        self: StepShiftedHurdleUncertainRF
            The model object.

        run_type : str
            The type of run to perform. Currently it is unlikely to affect the behaviour of the function.

        eval_type : str
            The type of evaluation to perform. Currently it is unlikely to affect the behaviour of the function.

        sequence_number : int
            The sequence number to predict outcomes for.


        Returns:
        --------
        pd.DataFrame
            The final predictions as a DataFrame.
        """

        final_preds = []  
        submodel_number = 0

        for submodel in tqdm(
            self._submodel_list, desc=f"Predicting submodel number", leave=True
        ):
            pred_by_step_binary = [
                self._predict_by_step(submodel[step][0], step, sequence_number)
                for step in self._steps
            ]
            pred_by_step_positive = [
                self._predict_by_step(submodel[step][1], step, sequence_number)
                for step in self._steps
            ]

            pred_concat_binary = pd.concat(pred_by_step_binary, axis=0)

            pred_concat_binary.rename(
                columns={f"pred_{self._targets}": "Classification"}, inplace=True
            )
            pred_concat_positive = pd.concat(pred_by_step_positive, axis=0)
            pred_concat_positive.rename(
                columns={f"pred_{self._targets}": "Regression"}, inplace=True
            )
            pred_concat = pd.concat([pred_concat_binary, pred_concat_positive], axis=1)
            pred_concat["submodel"] = submodel_number

            final_preds.append(pred_concat)
            submodel_number += 1

        final_preds_aslists = pd.concat(final_preds, axis=0)

        # Drawing samples from the classification model
        # Ensuring that the classification probabilities are between 0 and 1:
        final_preds_aslists["Classification"] = final_preds_aslists[
            "Classification"
        ].apply(lambda x: np.clip(x, 0, 1))
        final_preds_aslists["ClassificationSample"] = final_preds_aslists[
            "Classification"
        ].apply(lambda x: np.random.binomial(1, x, self._pred_samples))

        # Drawing samples from the regression model
        if self._log_target == True:
            if (
                self._draw_dist == "Poisson"
            ):  # Note: the Poisson distribution assumes a non-log-transformed target, so not defined here
                final_preds_aslists["RegressionSample"] = final_preds_aslists[
                    "Regression"
                ]
            if self._draw_dist == "Lognormal":
                # Draw from normal distribution for log-transformed outcomes, then exponentiate, then round to integer
                final_preds_aslists["RegressionSample"] = final_preds_aslists[
                    "Regression"
                ].apply(
                    lambda x: np.abs(
                        np.rint(
                            np.expm1(
                                np.random.normal(
                                    x, self._draw_sigma, self._pred_samples
                                )
                            )
                        )
                    )
                )

        if self._log_target == False:
            if (
                self._draw_dist == "Poisson"
            ):  # Note: this assumes a non-log-transformed target
                final_preds_aslists["RegressionSample"] = final_preds_aslists[
                    "Regression"
                ]
            if self._draw_dist == "Lognormal":
                final_preds_aslists["RegressionSample"] = final_preds_aslists[
                    "Regression"
                ].apply(
                    lambda x: np.abs(
                        np.rint(
                            np.expm1(
                                np.random.normal(
                                    np.log1p(x), self._draw_sigma, self._pred_samples
                                )
                            )
                        )
                    )
                )

        if self._draw_dist == "":
            final_preds_aslists["RegressionSample"] = final_preds_aslists["Regression"]

        # 'Explode' the samples to get one row per sample
        final_preds_full = final_preds_aslists.explode(
            ["ClassificationSample", "RegressionSample"]
        )
        final_preds_full["Prediction"] = (
            final_preds_full["ClassificationSample"]
            * final_preds_full["RegressionSample"]
        )

        # Ensuring that the final predictions are positive:
        final_preds_full["Prediction"] = final_preds_full["Prediction"].apply(
            lambda x: np.clip(x, 0, None)
        )

        # Column for the main prediction:
        pred_col_name = "pred_" + self._targets
        final_preds_full[pred_col_name] = final_preds_full["Prediction"]

        # Log-transforming the final predictions if the target is log-transformed, exponentiating if not, 
        # and adding a column with the log-transformed predictions
        if self._log_target == True:
            final_preds_full["LogPrediction"] = final_preds_full["Prediction"]
            final_preds_full["Prediction"] = np.expm1(final_preds_full["Prediction"])
        if self._log_target == False:
            final_preds_full["LogPrediction"] = np.log1p(final_preds_full["Prediction"])

        final_preds_full.drop(
            columns=[
                "Classification",
                "Regression",
                "ClassificationSample",
                "RegressionSample",
                "submodel",
                "Prediction",
                "LogPrediction",
            ],
            inplace=True,
        )
        final_preds = pd.DataFrame(
            final_preds_full.groupby(["month_id", "country_id"])[pred_col_name].apply(
                list
            )
        )

        return final_preds

    def predict(self, run_type: str, eval_type: str = "standard") -> pd.DataFrame:
        """
        Predicts outcomes based on the provided DataFrame and run type.

        Parameters:
        -----------
        df : pd.DataFrame
            The input data for making predictions.
        run_type : str
            The type of run to perform. If 'forecasting', a single prediction is made.
            Otherwise, multiple predictions are made based on the evaluation sequence number.
        eval_type : str, optional
            The type of evaluation to perform. Default is "standard".

        Returns:
        --------
        pd.DataFrame
            The final predictions as a DataFrame.
        """
        check_is_fitted(self, "is_fitted_")

        if run_type != "forecasting":
            preds = []
            if eval_type == "standard":
                for sequence_number in tqdm(
                    range(ForecastingModelManager._resolve_evaluation_sequence_number(eval_type)),
                    desc=f"Predicting for sequence number",
                    leave=True,
                ):
                    temp_preds_full = self.predict_sequence(sequence_number)
                    preds.append(temp_preds_full)
            else:
                raise ValueError(
                    f"{eval_type} is not supported now. Please use 'standard' evaluation type."
                )

        else:
            sequence_number = 0
            preds = self.predict_sequence(sequence_number)

        return preds



from views_stepshifter.models.darts_model import XGBRFClassifierModel, XGBRFModel
from views_stepshifter.models.hurdle_model import HurdleModel
from views_stepshifter.models.stepshifter import StepshifterModel

__all__ = [
    "XGBRFClassifierModel",
    "XGBRFModel",
    "HurdleModel",
    "StepshifterModel"
]


import functools
import numpy as np
import pandas as pd
import logging

logger = logging.getLogger(__name__)


def dataframe_is_right_format(dataframe: pd.DataFrame):
    pgm_cols = ["priogrid_gid", "month_id"]
    cm_cols = ["country_id", "month_id"]
    second_level_indices = ["priogrid_gid", "country_id", "priogrid_id"]
    try:
        assert isinstance(dataframe.index, pd.MultiIndex) and len(dataframe.index.levels) == 2
        # assert len(dataframe.index.levels) == 2
        # print("The dataframe has a two-level index")
    except AssertionError:
        logger.exception("Dataframe must have a two-level index")
        raise AssertionError("Dataframe must have a two-level index")
        
    try:
        # logger.info(f"First level of the index: {dataframe.index.names[0]}")
        assert dataframe.index.names[0] == "month_id"
        # print("The first level of the index is correct")
    except AssertionError:
        logger.exception("The first level of the index must be 'month_id'")
        raise AssertionError("The first level of the index must be 'month_id'")
    
    try:
        # logger.info(f"Second level of the index: {dataframe.index.names[1]}")
        assert dataframe.index.names[1] in second_level_indices
        # assert dataframe.index.names[1] in ["priogrid_gid"]
        # print("The second level of the index is correct")
    except AssertionError:
        logger.exception("The second level of the index must be 'country_id' or 'priogrid_gid'")
        raise AssertionError("The second level of the index must be 'country_id' or 'priogrid_gid'")

    try:
        assert set(dataframe.dtypes) == {np.dtype(float)}
        if "country_id" in dataframe.columns:
            assert set(dataframe.drop(columns=["country_id"]).dtypes) == {np.dtype(float)}
        else:
            assert set(dataframe.dtypes) == {np.dtype(float)}
        # print("The dataframe contains only np.float64 floats")
    except AssertionError:
        logger.exception("The dataframe must contain only np.float64 floats")
        raise AssertionError("The dataframe must contain only np.float64 floats")
    
# Needs update
def views_validate(fn):
    @functools.wraps(fn)
    def inner(*args, **kwargs):
        # When fit()
        #args[0] will be the instance of the StepshifterModel class (self).
        #args[1] will be the DataFrame (df).

        # When predict()
        # args[0] will be the instance of the StepshifterModel class (self).
        # args[1] will be the DataFrame (df).
        # args[2] will be the string (run_type).
        # args[3] will be the optional string (eval_type), if provided.
        dataframe = args[0] if isinstance(args[0], pd.DataFrame) else args[1] # Hardcoding this makes me uncomfortable
        dataframe_is_right_format(dataframe)
        # dataframe_is_right_format(args[1]) # is also uncomfortable
        return fn(*args, **kwargs)
    return inner


from views_pipeline_core.managers.model import ForecastingModelManager
from views_stepshifter.models.stepshifter import StepshifterModel
from views_stepshifter.models.validation import views_validate
from sklearn.utils.validation import check_is_fitted
import pandas as pd
from typing import List, Dict
import logging
import tqdm
from concurrent.futures import ProcessPoolExecutor, as_completed
from functools import partial
logger = logging.getLogger(__name__)


class HurdleModel(StepshifterModel):
    """
    Hurdle model for time series forecasting. The model consists of two stages:
    1. Binary stage: Predicts whether the target variable is 0 or > 0.
    2. Positive stage: Predicts the value of the target variable when it is > 0.

    Note:
    This algorithm uses a two-step approach.

    **Step 1: Classification Stage**
    In the first step, a classification model is used with a binary target (0 or 1),
    indicating the absence or presence of violence.

    **Step 2: Regression Stage**
    In the second step, we use a regression model to predict a continuous or count value
    (e.g., the expected number of conflict fatalities) for the selected time series.
    We include the entire time series for countries or PRIO grids where the
    classification stage yielded at least one "positive" prediction,
    rather than limiting the regression to just the predicted positive values.
    """

    def __init__(self, config: Dict, partitioner_dict: Dict[str, List[int]]):
        super().__init__(config, partitioner_dict)
        self._clf_params = self._get_parameters(config)["clf"]
        self._reg_params = self._get_parameters(config)["reg"]

    def _resolve_clf_model(self, func_name: str):
        """Lookup table for supported classification models"""

        match func_name:
            case "XGBClassifier":
                from darts.models import XGBClassifierModel
                # if self.get_device_params().get("device") == "cuda":
                #     logger.info("\033[92mUsing CUDA for XGBClassifierModel\033[0m")
                #     cuda_params = {"tree_method": "hist", "device": "cuda"}
                #     return partial(XGBClassifierModel, **cuda_params)
                return XGBClassifierModel
            case "XGBRFClassifier":
                from views_stepshifter.models.darts_model import XGBRFClassifierModel
                # if self.get_device_params().get("device") == "cuda":
                #     logger.info("\033[92mUsing CUDA for XGBRFClassifierModel\033[0m")
                #     cuda_params = {"tree_method": "hist", "device": "cuda"}
                #     return partial(XGBRFClassifierModel, **cuda_params)
                return XGBRFClassifierModel
            case "LGBMClassifier":
                from darts.models import LightGBMClassifierModel
                return LightGBMClassifierModel
            case _:
                raise ValueError(
                    f"Model {func_name} is not a valid Darts forecasting model or is not supported now. "
                    f"Change the model in the config file."
                )

    def _fit_by_step(self, step, target_binary, target_pos, past_cov_pos):
        # Fit binary-like stage using a classification model
        binary_model = self._clf(lags_past_covariates=[-step], **self._clf_params)
        binary_model.fit(target_binary, past_covariates=self._past_cov)

        # Fit positive stage using the regression model
        positive_model = self._reg(lags_past_covariates=[-step], **self._reg_params)
        positive_model.fit(target_pos, past_covariates=past_cov_pos)

        return (binary_model, positive_model)
    
    def _predict_by_sequence(self, sequence_number):
        pred_by_step_binary = []
        pred_by_step_positive = []
        
        for step in self._steps:
            # Predict for binary model
            pred_binary = self._predict_by_step(self._models[step][0], step, sequence_number)
            pred_by_step_binary.append(pred_binary)
            
            # Predict for positive model
            pred_positive = self._predict_by_step(self._models[step][1], step, sequence_number)
            pred_by_step_positive.append(pred_positive)
        
        final_pred = (
            pd.concat(pred_by_step_binary, axis=0).sort_index() *
            pd.concat(pred_by_step_positive, axis=0).sort_index()
        )
    
        return final_pred

    @views_validate
    def fit(self, df: pd.DataFrame):
        df = self._process_data(df)
        self._prepare_time_series(df)
        self._clf = self._resolve_clf_model(self._config["model_clf"])
        self._reg = self._resolve_reg_model(self._config["model_reg"])

        # Binary outcome (event/no-event)
        # According to the DARTS doc, if timeseries uses a numeric type different from np.float32 or np.float64, not all functionalities may work properly.
        # So use astype(float) instead of astype(int) (we should have binary outputs 0,1 though)
        target_binary = [
            s.map(lambda x: (x > 0).astype(float)) for s in self._target_train
        ]

        # Positive outcome (for cases where target > 0)
        target_pos, past_cov_pos = zip(
            *[
                (t, p)
                for t, p in zip(self._target_train, self._past_cov)
                if (t.values() > 0).any()
            ]
        )

        if self.get_device_params().get("device") == "cuda":
            for step in tqdm.tqdm(self._steps, desc="Fitting model for step", leave=True):
                # Fit binary-like stage using a classification model, but the target is binary (0 or 1)
                binary_model = self._clf(lags_past_covariates=[-step], **self._clf_params)
                binary_model.fit(target_binary, past_covariates=self._past_cov)

                # Fit positive stage using the regression model
                positive_model = self._reg(lags_past_covariates=[-step], **self._reg_params)
                positive_model.fit(target_pos, past_covariates=past_cov_pos)
                self._models[step] = (binary_model, positive_model)
        else:
            models = {}
            with ProcessPoolExecutor() as executor:
                futures = {
                    executor.submit(self._fit_by_step, step, target_binary, target_pos, past_cov_pos): step
                    for step in self._steps
                }
                for future in tqdm.tqdm(as_completed(futures.keys()), desc="Fitting models for steps", total=len(futures)):
                    step = futures[future]
                    models[step] = future.result()
            self._models = models
        self.is_fitted_ = True

    def predict(self, run_type: str, eval_type: str = "standard") -> pd.DataFrame:
        check_is_fitted(self, "is_fitted_")

        if run_type != "forecasting":

            if eval_type == "standard":
                total_sequence_number = (
                    ForecastingModelManager._resolve_evaluation_sequence_number(eval_type)
                )
                if self.get_device_params().get("device") == "cuda":
                    preds = []
                    for sequence_number in tqdm.tqdm(
                        range(ForecastingModelManager._resolve_evaluation_sequence_number(eval_type)),
                        desc="Predicting for sequence number",
                    ):
                        pred_by_step_binary = [
                            self._predict_by_step(
                                self._models[step][0], step, sequence_number
                            )
                            for step in self._steps
                        ]
                        pred_by_step_positive = [
                            self._predict_by_step(
                                self._models[step][1], step, sequence_number
                            )
                            for step in self._steps
                        ]
                        pred = pd.concat(pred_by_step_binary, axis=0) * pd.concat(pred_by_step_positive, axis=0)
                        preds.append(pred)

                else:
                    preds = [None] * total_sequence_number
                    with ProcessPoolExecutor() as executor:
                        futures = {
                            executor.submit(self._predict_by_sequence, sequence_number): sequence_number
                            for sequence_number in range(ForecastingModelManager._resolve_evaluation_sequence_number(eval_type))
                        }
                        for future in tqdm.tqdm(
                            as_completed(futures.keys()),
                            desc="Predicting for sequence number",
                            total=len(futures),
                        ):
                            sequence_number = futures[future]
                            preds[sequence_number] = future.result()
            else:
                raise ValueError(
                    f"{eval_type} is not supported now. Please use 'standard' evaluation type."
                )

        else:
            if self.get_device_params().get("device") == "cuda":
                pred_by_step_binary = []
                pred_by_step_positive = []
                for step in tqdm.tqdm(self._steps, desc="Predicting for step", total=len(self._steps)):
                    pred_by_step_binary.append(
                        self._predict_by_step(self._models[step][0], step, 0)
                    )
                    pred_by_step_positive.append(
                        self._predict_by_step(self._models[step][1], step, 0)
                    )
                
                preds = pd.concat(pred_by_step_binary, axis=0) * pd.concat(
                    pred_by_step_positive, axis=0
                )
  
            else:
                with ProcessPoolExecutor() as executor:
                    futures_binary = {
                        step: executor.submit(
                            self._predict_by_step, self._models[step][0], step, 0
                        )
                        for step in self._steps
                    }
                    futures_positive = {
                        step: executor.submit(
                            self._predict_by_step, self._models[step][1], step, 0
                        )
                        for step in self._steps
                    }

                    pred_by_step_binary = [
                        future.result()
                        for future in tqdm.tqdm(
                            as_completed(futures_binary.values()),
                            desc="Predicting binary outcomes",
                            total=len(futures_binary),
                        )
                    ]
                    pred_by_step_positive = [
                        future.result()
                        for future in tqdm.tqdm(
                            as_completed(futures_positive.values()),
                            desc="Predicting positive outcomes",
                            total=len(futures_positive),
                        )
                    ]

                    preds = (
                        pd.concat(pred_by_step_binary, axis=0).sort_index()
                        * pd.concat(pred_by_step_positive, axis=0).sort_index()
                    )
        return preds



import xgboost as xgb
from darts.models import XGBModel, XGBClassifierModel

class XGBRFModel(XGBModel):
    """
    A Darts wrapper for the XGBoost Random Forest Regressor.
    
    This class inherits all functionality from darts.models.XGBModel
    and simply replaces the underlying model with 
    `xgb.XGBRFRegressor`.
    """

    @staticmethod
    def _create_model(**kwargs):
        """Create an XGBRFRegressor model."""
        return xgb.XGBRFRegressor(**kwargs)

class XGBRFClassifierModel(XGBClassifierModel):
    """
    A Darts wrapper for the XGBoost Random Forest Classifier.
    
    This class inherits all functionality from darts.models.XGBClassifierModel
    and simply replaces the underlying model with 
    `xgb.XGBRFClassifier`.
    """

    @staticmethod
    def _create_model(**kwargs):
        """Create an XGBRFClassifier model."""
        return xgb.XGBRFClassifier(**kwargs)


import pickle
import numpy as np
import pandas as pd
import logging
from darts import TimeSeries
from sklearn.utils.validation import check_is_fitted
from typing import List, Dict
from views_stepshifter.models.validation import views_validate
from views_pipeline_core.managers.model import ForecastingModelManager
import tqdm
from concurrent.futures import ProcessPoolExecutor, as_completed
import torch
from functools import partial

logger = logging.getLogger(__name__)


class StepshifterModel:
    def __init__(self, config: Dict, partitioner_dict: Dict[str, List[int]]):
        self._config = config
        self._steps = config["steps"]
        self._reg_params = self._get_parameters(config)
        self._train_start, self._train_end = partitioner_dict["train"]
        self._test_start, self._test_end = partitioner_dict["test"]
        self._models = {}

        # Multiple targets handling
        if not isinstance(config["targets"], list):
            raise ValueError("Dependent variable must be a list")
        elif len(config["targets"]) > 1:
            raise ValueError("Stepshifter only supports one dependent variable")
        else:
            self._targets = config["targets"][0]

    @staticmethod
    def get_device_params():
        if torch.cuda.is_available():
            return {"device": "cuda"}
        elif torch.backends.mps.is_available():
            return {"device": "mps"}
        else:
            return {}

    def _resolve_reg_model(self, func_name: str):
        """
        Lookup table for supported regression models
        Note that stepshifter doesn't support cuda for now before we figure out how to move the input data to the GPU. (fit works intelligently but predict doesn't)
        """

        match func_name:
            case "XGBRFRegressor":
                from views_stepshifter.models.darts_model import XGBRFModel
                # if self.get_device_params().get("device") == "cuda":
                #     logger.info("\033[92mUsing CUDA for XGBRFRegressor\033[0m")
                #     cuda_params = {"tree_method": "hist", "device": "cuda"}
                #     return partial(XGBRFModel, **cuda_params)
                return XGBRFModel
            case "XGBRegressor":
                from darts.models import XGBModel
                # if self.get_device_params().get("device") == "cuda":
                #     logger.info("\033[92mUsing CUDA for XGBRegressor\033[0m")
                #     cuda_params = {"tree_method": "hist", "device": "cuda"}
                #     return partial(XGBModel, **cuda_params)
                return XGBModel
            case "LGBMRegressor":
                from darts.models import LightGBMModel
                # if self.get_device_params().get("device") == "cuda":
                #     logger.info("\033[92mUsing CUDA for LGBMRegressor\033[0m")
                #     cuda_params = {"device": "cuda"}
                #     return partial(LightGBMModel, **cuda_params)
                return LightGBMModel
            case _:
                raise ValueError(
                    f"Model {func_name} is not a valid forecasting model or is not supported now. "
                    f"Change the model in the config file."
                )

    def _get_parameters(self, config: Dict):
        """
        Get the parameters from the config file.
        If not sweep, then get directly from the config file, otherwise have to remove some parameters.
        """

        if config["sweep"]:
            parameters = {k: v for k, v in config.items() if k in ["clf", "reg"]}
        else:
            parameters = config["parameters"]

        return parameters

    def _process_data(self, df: pd.DataFrame):
        """
        Countries appear and disappear, so we are predicting countries that exist in the last month of the training data.
        If the country appeared earlier but don't have data previously, we will fill the missing data with 0.
        """

        # set up
        self._time = df.index.names[0]
        self._level = df.index.names[1]
        self._independent_variables = [c for c in df.columns if c != self._targets]

        last_month_id = df.index.get_level_values(self._time).max()
        existing_country_ids = df.loc[last_month_id].index.unique()
        df = df[df.index.get_level_values(self._level).isin(existing_country_ids)]

        all_months = df.index.get_level_values(self._time).unique()
        all_combinations = pd.MultiIndex.from_product(
            [all_months, existing_country_ids], names=[self._time, self._level]
        )
        missing_combinations = all_combinations.difference(df.index)

        missing_df = pd.DataFrame(0, index=missing_combinations, columns=df.columns)
        df = pd.concat([df, missing_df]).sort_index()

        df[self._targets] = np.log1p(df[self._targets]) # Calculates log(1 + x).
        return df

    def _prepare_time_series(self, df: pd.DataFrame):
        """
        Prepare time series for training and prediction
        """

        df_reset = df.reset_index(level=[1])
        self._series = TimeSeries.from_group_dataframe(
            df_reset,
            group_cols=self._level,
            value_cols=self._independent_variables + [self._targets],
        )

        self._target_train = [
            series.slice(self._train_start, self._train_end + 1)[self._targets]
            for series in self._series
        ]  # ts.slice is different from df.slice
        self._past_cov = [
            series[self._independent_variables] for series in self._series
        ]

    def _fit_by_step(self, step):
        model = self._reg(lags_past_covariates=[-step], **self._reg_params)
        model.fit(self._target_train, past_covariates=self._past_cov)
        return model

    def _predict_by_step(self, model, step: int, sequence_number: int):
        """
        Keep predictions with last-month-with-data, i.e., diagonal prediction
        """
        # logger.info(f"Starting prediction for step: {step}")
        target = [
            series.slice(self._train_start, self._train_end + 1 + sequence_number)[
                self._targets
            ]
            for series in self._series
        ]
        ts_pred = model.predict(
            n=step,
            series=target,
            # darts automatically locates the time period of past_covariates
            past_covariates=self._past_cov,
            show_warnings=False,
        )

        # process the predictions
        index_tuples, df_list = [], []
        for pred in ts_pred:
            df_pred = pred.to_dataframe().loc[
                [self._test_start + step + sequence_number - 1]
            ]
            level = int(pred.static_covariates.iat[0, 0])
            index_tuples.extend([(month, level) for month in df_pred.index])
            df_list.append(df_pred.values)

        df_preds = pd.DataFrame(
            data=np.concatenate(df_list),
            index=pd.MultiIndex.from_tuples(
                index_tuples, names=[self._time, self._level]
            ),
            columns=[f"pred_{self._targets}"],
        )

        df_preds[f"pred_{self._targets}"] = np.expm1(df_preds[f"pred_{self._targets}"]) # Calculates exp(x) - 1
        
        return df_preds.sort_index()

    def _predict_by_sequence(self, sequence_number):
        pred_by_step = []
        for step in self._steps:
            pred = self._predict_by_step(self._models[step], step, sequence_number)
            pred_by_step.append(pred)
        return pd.concat(pred_by_step, axis=0).sort_index()

    # @views_validate
    # def fit(self, df: pd.DataFrame):
    #     df = self._process_data(df)
    #     self._prepare_time_series(df)
    #     self._reg = self._resolve_reg_model(self._config["model_reg"])
    #     for step in tqdm.tqdm(
    #         self._steps, desc="Fitting model for step", leave=True
    #     ):
    #         model = self._reg(lags_past_covariates=[-step], **self._reg_params)
    #         model.fit(self._target_train,
    #                   past_covariates=self._past_cov) # Darts will automatically ignore the parts of past_covariates that go beyond the training period
    #         self._models[step] = model
    #     self.is_fitted_ = True

    @views_validate
    def fit(self, df: pd.DataFrame):
        df = self._process_data(df)
        self._prepare_time_series(df)
        self._reg = self._resolve_reg_model(self._config["model_reg"])

        models = {}
        if self.get_device_params().get("device") == "cuda":
            for step in tqdm.tqdm(
                self._steps, desc="Fitting model for step", leave=True
            ):
                model = self._reg(lags_past_covariates=[-step], **self._reg_params)
                model.fit(self._target_train,
                            past_covariates=self._past_cov) # Darts will automatically ignore the parts of past_covariates that go beyond the training period
                self._models[step] = model
        else:
            with ProcessPoolExecutor() as executor:
                futures = {
                    executor.submit(self._fit_by_step, step): step for step in self._steps
                }
                for future in tqdm.tqdm(
                    futures.keys(), desc="Fitting models for steps", total=len(futures)
                ):
                    step = futures[future]
                    models[step] = future.result()
                self._models = models  # Use local variable to avoid concurrent execution issues
        self.is_fitted_ = True

    def predict(self, run_type: str, eval_type: str = "standard") -> pd.DataFrame:
        check_is_fitted(self, "is_fitted_")

        if run_type != "forecasting":

            if eval_type == "standard":
                total_sequence_number = (
                    ForecastingModelManager._resolve_evaluation_sequence_number(eval_type)
                )

                if self.get_device_params().get("device") == "cuda":
                    preds = []
                    for sequence_number in tqdm.tqdm(
                        range(ForecastingModelManager._resolve_evaluation_sequence_number(eval_type)),
                        desc="Predicting for sequence number",
                    ):
                        pred_by_step = [
                            self._predict_by_step(self._models[step], step, sequence_number)
                            for step in self._steps
                        ]
                        pred = pd.concat(pred_by_step, axis=0)
                        preds.append(pred)
                else:
                    preds = [None] * total_sequence_number
                    with ProcessPoolExecutor() as executor:
                        futures = {
                            executor.submit(
                                self._predict_by_sequence, sequence_number
                            ): sequence_number
                            for sequence_number in range(total_sequence_number)
                        }

                        for future in tqdm.tqdm(
                            as_completed(futures.keys()),
                            desc="Predicting for sequence number",
                            total=len(futures),
                        ):
                            sequence_number = futures[future]
                            preds[sequence_number] = future.result()
            else:
                raise ValueError(
                    f"{eval_type} is not supported now. Please use 'standard' evaluation type."
                )

        else:

            if self.get_device_params().get("device") == "cuda":
                preds = []
                for step in tqdm.tqdm(self._steps, desc="Predicting for steps"):
                    preds.append(self._predict_by_step(self._models[step], step, 0))
                preds = pd.concat(preds, axis=0).sort_index()
                
            else:
                with ProcessPoolExecutor() as executor:
                    futures = {
                        step: executor.submit(
                            self._predict_by_step, self._models[step], step, 0
                        )
                        for step in self._steps
                    }
                    preds_by_step = [
                        future.result()
                        for future in tqdm.tqdm(
                            as_completed(futures.values()),
                            desc="Predicting outcomes",
                            total=len(futures),
                        )
                    ]

                preds = pd.concat(preds_by_step, axis=0).sort_index()

        return preds

    def save(self, path: str):
        try:
            with open(path, "wb") as file:
                pickle.dump(self, file)
            logger.info(f"Model successfully saved to {path}")
        except Exception as e:
            logger.exception(f"Failed to save model: {e}")

    @property
    def models(self):
        return list(self._models.values())

    @property
    def steps(self):
        return self._steps

    @property
    def targets(self):
        return self._targets



from multiprocessing import set_start_method, get_start_method
from views_stepshifter.manager.stepshifter_manager import StepshifterManager

try:
    set_start_method('spawn')
    print(f"Multiprocessing start method set to: {get_start_method()}") # Use print because logger is not yet configured
except RuntimeError:
    # print(f"Multiprocessing start method is already set to: {get_start_method()}") 
    pass

__all__ = ["StepshifterManager"]